#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <dbus/dbus.h>
#include <sane/sane.h>

#include "imcad.h"
#include "dbus-methods.h"
#include "jobs.h"

//table of dbus methods with functions to handle them
void *methodtable[] =
	{"GetDeviceList", method_get_device_list, "OpenDevice", method_open_device,
	"StartScan", method_start_scan,	"StopScan", method_stop_scan,
	"ReadImage", method_read_image, "GetOptionInfo", method_get_option_info,
	"GetNumOptions", method_get_num_options, "SetOptionValue", method_set_option_value,
	"GetOptionValue", method_get_option_value, NULL};
//used to stop main loop on signal
int run = 1;


void handle_signal (int signal) {
	//if we caught a SIGINT or SIGTERM, stop the main loop
	printf("Exiting...\n");
	run = 0;
}


DBusHandlerResult handle_request (DBusConnection *conn, DBusMessage *msg, void *userdata) {
	DBusMessage *reply;

	//look through the method table to see if we can handle this message
	int i = 0;
	while (TRUE) {
		//make sure we're not at the end of the table
		if (methodtable[i] == NULL) {
			printf("Got unrecognized method call\n");
			return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
		//if not, check if this call matches the received one
		} else if (dbus_message_is_method_call(msg,
			"org.freedesktop.ImageCaptureInterface", methodtable[i])) {

			//if it does, call the appropriate method
			printf("Method %s called\n", (char *)methodtable[i]);
			//but first, create a reply message for it to use
			reply = dbus_message_new_method_return(msg);
			void (*callback)(DBusMessage*, DBusMessage*) =
				(void (*)(DBusMessage*, DBusMessage*))methodtable[++i];
			//TODO: create new thread for function so long I/O operations don't block requests
			// will have to be careful about jobs timing out during read (reset in read loop?)
			callback(msg, reply);
			//once the method's done, send the reply
			dbus_connection_send(conn, reply, NULL);
			dbus_connection_flush(conn);
			dbus_message_unref(reply);
			break;
		}
		//if it doesn't match, look at the next one
		i += 2;
	}
	return DBUS_HANDLER_RESULT_HANDLED;
}


int main (int argc, char **argv) {

	DBusConnection *conn;
	DBusError error;

	//init SANE
	if (sane_init(NULL, NULL) != SANE_STATUS_GOOD) {
		printf("Couldn't init SANE!\n");
		exit(1);
	}

	dbus_error_init(&error);
	//connect to DBUS
	conn = dbus_bus_get(DBUS_BUS_SESSION, &error);
	if (!conn) {
		printf("Couldn't connect to DBUS!\n");
		if (dbus_error_is_set(&error)) {
			printf("%s\n", error.message);
			dbus_error_free(&error);
		}
		sane_exit();
		exit(1);
	}

	//try to get a name
	if (dbus_bus_request_name(conn, "org.freedesktop.ImageCapture",
		DBUS_NAME_FLAG_REPLACE_EXISTING, &error) !=
		DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER) {

		printf("Couldn't lock name!\n");
		if (dbus_error_is_set(&error)) {
			printf("%s\n", error.message);
			dbus_error_free(&error);
		}
		dbus_connection_unref(conn);
		sane_exit();
		exit(1);
	}

	//export an object for requests
	DBusObjectPathVTable vtable;
	vtable.message_function = handle_request;
	vtable.unregister_function = NULL;
	if (!dbus_connection_register_object_path(conn,
		"/org/freedesktop/ImageCapture", &vtable, NULL)) {

		printf("Couldn't export object!\n");
		dbus_connection_unref(conn);
		sane_exit();
		exit(1);
	}

	//install a signal handler so the daemon can exit cleanly,
	struct sigaction sighandler;
	sighandler.sa_handler = handle_signal;
	if (sigaction(SIGINT, &sighandler, NULL) == -1 ||
		sigaction(SIGTERM, &sighandler, NULL) == -1) {

		printf("Couldn't install signal handler, continuing...\n");
	}

	//and do nothing until we get an exit signal
	while (dbus_connection_read_write_dispatch(conn, SIG_CHECK_INT) && run) {
		//...except for waking up to check if any job handles have timed out
		check_job_timeouts();
	}

	dbus_connection_unref(conn);
	sane_exit();
	return 0;
}
